#ifdef WIN32
#include <windows.h>
#endif
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

 
#ifdef dDOUBLE
  #define dsDrawBox dsDrawBoxD
  #define dsDrawSphere dsDrawSphereD
  #define dsDrawLine dsDrawLineD
#endif

#define MAX_CONTACTS 4 // ő̏Փˌo\

static dWorldID world;
static dSpaceID space;
static dJointGroupID contactgroup;

static dBodyID body_box;
static dGeomID geom_box;
static dGeomID geom_ray;

static dBodyID body_ball;
static dGeomID geom_ball;

dReal  dist; // rayɐڐĜ̋ێ


// Փˌop֐
static void nearCallback( void *data, dGeomID o1, dGeomID o2 )
{
  dBodyID b1 = dGeomGetBody( o1 ); // 1
  dBodyID b2 = dGeomGetBody( o2 ); // 2
 
  if ( b1 && b2 && dAreConnectedExcluding( b1, b2, dJointTypeContact ) )
    return; // ՓˑΏۂłȂ̂̏Փ˂͂͂

  //////////////////////////////////////////////////////////
  // 擾
  if (o1==geom_ray || o2==geom_ray) {
      dContactGeom c;
      int numc = dCollide( o1, o2, 1, &c, sizeof( dContact ) );
      if (numc > 0) {
          dist = c.depth; // depthray̎n_ʒű
      }
      return;
  }
  //////////////////////////////////////////////////////////
 
  dContact contact[MAX_CONTACTS];
  for ( int i=0; i<MAX_CONTACTS; i++ )
  {
    // ̓m̐ڐG̃p[^ݒ
    contact[i].surface.mode = dContactBounce | dContactSoftERP | dContactSoftCFM;
    contact[i].surface.mu = dInfinity;     // CW
    contact[i].surface.bounce = 0.5;       // W
    contact[i].surface.soft_erp = 0.3;     // ERPݒ
    contact[i].surface.soft_cfm = 0.0001;  // CFMݒ
  }
 
  // Փˌo
  int numc = dCollide( o1, o2, MAX_CONTACTS, &contact[0].geom, sizeof( dContact ) ); 
  if ( numc > 0 )
  {
    for ( int i=0; i<numc; i++ )
    {
      // Փ˂̔
      dJointID c = dJointCreateContact( world, contactgroup, contact+i );
                                 // ڐGWCg
      dJointAttach( c, b1, b2 ); // ڐGWCgڑ
    }
  }
}
 
// start simulation - set viewpoint
static void start()
{
  static float xyz[3] = { 2.f, -10.f, 3.f };
  static float hpr[3] = { 90.f, -15.f, 0.f };
 
  dsSetViewpoint( xyz, hpr );
}
 
// simulation loop
static void simLoop( int pause )
{
    dist = -1.0;

    // Ctl+p ꂽifɓȂ
    if (!pause)
    {
        dSpaceCollide( space, 0, &nearCallback ); // Փˌo

        dWorldStep( world, 0.01 );
        
        dJointGroupEmpty( contactgroup );
    }

    ///////////////////////////////////////////////////
    // \
    if (dist>=0.0) {
        printf( "dist = %f\n", dist );
    }
    ///////////////////////////////////////////////////
 
    // draw box
    {
        dsSetColor( 1.0f, 1.0f, 0.0f );
        dVector3 size;
        dGeomBoxGetLengths( geom_box, size );
        dsDrawBox( dGeomGetPosition( geom_box ), dGeomGetRotation( geom_box ), size );
    }
    // draw ray
    {
        dsSetColor( 1.0f, 0.0f, 0.0f );

        dVector3 Origin, Direction;
        dGeomRayGet(geom_ray, Origin, Direction);
        dReal Length = dGeomRayGetLength(geom_ray);

        dVector3 End;
        End[0] = Origin[0] + (Direction[0] * Length);
        End[1] = Origin[1] + (Direction[1] * Length);
        End[2] = Origin[2] + (Direction[2] * Length);
        End[3] = Origin[3] + (Direction[3] * Length);

        dsDrawLine(Origin, End);
    }
    // draw ball
    {
        dsSetColor( 0.0f, 1.0f, 1.0f );
        dReal radius = dGeomSphereGetRadius( geom_ball );
        dsDrawSphere( dGeomGetPosition( geom_ball ), dGeomGetRotation( geom_ball ), radius );
    }
}
 
 
int main( int argc, char* argv[] )
{
    dInitODE();
 
    // setup pointers to drawstuff callback functions
    dsFunctions fn;
    fn.version = DS_VERSION;
    fn.start   = &start;
    fn.step    = &simLoop;
    fn.command = 0;
    fn.stop    = 0;
    fn.path_to_textures = "../drawstuff/textures";
 
    world = dWorldCreate();
    dWorldSetGravity( world, 0.0, 0.0, -9.8 );

    space = dHashSpaceCreate( 0 ); 
    contactgroup = dJointGroupCreate( 0 );

    // plane create
    {
        dCreatePlane( space, 0, 0, 1, 0 ); 
    }

    // box creating
    {
        dReal pos[3] = { 0.0, 0.0, 1.5 };
        dReal size[3] = { 2.0, 2.0, 2.0 };

        // body setting
        body_box = dBodyCreate( world );
        dBodySetPosition( body_box, pos[0], pos[1], pos[2] );
        // geom setting
        geom_box = dCreateBox( space, size[0], size[1], size[2] );
        dGeomSetBody( geom_box, body_box );
        // mass setting
        dMass mass;
        dMassSetBoxTotal( &mass, 10.0, size[0], size[1], size[2] ); // 10kg
        dBodySetMass( body_box, &mass );
        // rotation
        dMatrix3 R;
        dRFromAxisAndAngle( R, 1.0, 0.0, 0.0, 0.0*M_PI/180 );
        dBodySetRotation( body_box, R );

        //// fixed joint setting
        //dJointID fixed;
        //fixed = dJointCreateFixed( world, 0 );
        //dJointAttach( fixed, NULL, body_box );
        //dJointSetFixed( fixed );
    }

    // Ray setting
    {
        dReal pos[3] = { 1.0, 0.0, 0.0 };
        dReal length = 5.0;

        geom_ray = dCreateRay( space, length );
        dGeomSetBody( geom_ray, body_box );  // <= body is box

        // position and rotation setting
        dGeomSetOffsetPosition( geom_ray, pos[0], pos[1], pos[2] );
        dMatrix3 R;
        dRFromAxisAndAngle( R, 0.0, 1.0, 0.0, 90.0*M_PI/180 );
        dGeomSetOffsetRotation( geom_ray, R );
        //dGeomRaySet( geom_ray, pos[0], pos[1], pos[2], dir[0], dir[1], dir[2] );
    }
    
    // ball creating
    {
        dReal pos[3] = { 10.0, 0.0, 1.0 };
        dReal radius = 1.0;

        // body setting
        body_ball = dBodyCreate( world );
        dBodySetPosition( body_ball, pos[0], pos[1], pos[2] );
        // geom setting
        geom_ball = dCreateSphere( space, radius );
        dGeomSetBody( geom_ball, body_ball );
        // mass setting
        dMass mass;
        dMassSetSphereTotal( &mass, 1.0, radius ); // 10kg
        dBodySetMass( body_ball, &mass );
        // rotation
        dMatrix3 R;
        dRFromAxisAndAngle( R, 1.0, 0.0, 0.0, 0.0*M_PI/180 );
        dBodySetRotation( body_ball, R );
    }

    //{[]
    dBodyAddForce( body_ball, -500.0, 0.0, 0.0 );


    dsSimulationLoop( argc, argv, 320, 240, &fn );

    dWorldDestroy( world );
    dCloseODE();
    return 0;
}



